#include "ITPResource.h"

extern char* MyName;
extern char* nw_inp_dir;

ITPResource::ITPResource() : 
    itpFileName(""),offset(0),listp(NULL),datap(NULL),tagp(NULL),
    valuep(NULL),rootp(NULL),indexp(NULL) {
}

ITPResource::ITPResource(std::string fname, int o) : 
    itpFileName(fname),offset(o),listp(NULL),datap(NULL),tagp(NULL),
    valuep(NULL),rootp(NULL),indexp(NULL) {

    printf("constructing ITP resource at %d\n", offset);
    readITPFile();
}

ITPResource& ITPResource::fromFile(std::string fname, int o) {
    itpFileName = fname;
    offset = o;
    readITPFile();
    return *this;
}

ITPResource::~ITPResource() {
    if(listp) {
	delete listp;
    }
    if(datap) {
	delete datap;
    }
    if(tagp) {
	delete tagp;
    }
    if(valuep) {
	delete valuep;
    }
    if(rootp) {
	delete rootp;
    }
    if(indexp) {
	delete indexp;
    }
}

bool ITPResource::checkSignature(const char* s, const char* v) const {
    /* this format is used for several resources; match the ones we can
    **	cope with
    */
    if (strncmp(s, NWN_ITP_SIGNATURE, sizeof s) == 0 &&
	strncmp(v, NWN_ITP_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, ARE_SIGNATURE, sizeof s) == 0 &&
	    strncmp(v, ARE_VERSION, sizeof v) == 0)
	    ;
    else if (strncmp(s, UTI_SIGNATURE, sizeof s) == 0 &&
		strncmp(v, UTI_VERSION, sizeof v) == 0)
		;
    else if (strncmp(s, UTC_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, UTC_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, GIT_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, GIT_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, GIC_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, GIC_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, UTP_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, UTP_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, UTE_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, UTE_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, UTD_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, UTD_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, UTS_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, UTS_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, UTM_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, UTM_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, JRL_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, JRL_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, IFO_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, IFO_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, FAC_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, FAC_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, DLG_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, DLG_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, UTW_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, UTW_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, BIC_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, BIC_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, GUI_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, GUI_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, UTT_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, UTT_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, PTT_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, PTT_VERSION, sizeof v) == 0)
	;
    else if (strncmp(s, GFF_SIGNATURE, sizeof s) == 0 &&
	     strncmp(v, GFF_VERSION, sizeof v) == 0)
	;
    else {
	fprintf(stderr, "%s: Invalid signature <%-.4s> or version <%-.4s>!\n",
		MyName, s, v);
	return false;
    }
    return true;
}

void ITPResource::readITPFile() {
    FILE* itpFile = fopen(itpFileName.c_str(),"rb");
    fseek(itpFile,offset,SEEK_SET);

    if (fread(&head, sizeof head, 1, itpFile) != 1) {
	fprintf(stderr, "%s: Can't read ITP head!\n", MyName);
	return;
    }
    if(!checkSignature(head.signature,head.version)) {
	return;
    }
#if WORDS_BIGENDIAN
    be2lel(&head.oList);
    be2lel(&head.cList);
    be2lel(&head.oData);
    be2lel(&head.cData);
    be2lel(&head.oTag);
    be2lel(&head.cTag);
    be2lel(&head.oValue);
    be2lel(&head.cValue);
    be2lel(&head.oRoot);
    be2lel(&head.cRoot);
    be2lel(&head.oIndex);
    be2lel(&head.cIndex);
#endif
	    
    listp = new nwn_itp_list[head.cList];
    if (!listp)
    {
	fprintf(stderr, "%s: Can't allocate %ld list's!\n",
		MyName, head.cList);
	return;
    }
    fseek(itpFile,  offset + head.oList, SEEK_SET);
    if (fread(listp, sizeof *listp, head.cList, itpFile) != head.cList)
    {
	fprintf(stderr, "%s: Can't read %ld list's!\n", MyName, head.cList);
	return;
    }
#if WORDS_BIGENDIAN
    for (int i = 0; i < head.cList; ++i)
    {
	be2lel(&listp[i].code);
	be2lel(&listp[i].index);
	be2lel(&listp[i].count);
    }
#endif
	    
    datap = new nwn_itp_data[head.cData];
    if (!datap)
    {
	fprintf(stderr, "%s: Can't allocate %ld data's!\n",
		MyName, head.cData);
	return;
    }
    fseek(itpFile, offset + head.oData, SEEK_SET);
    if (fread(datap, sizeof *datap, head.cData, itpFile) != head.cData)
    {
	fprintf(stderr, "%s: Can't read %ld data's!\n", MyName, head.cData);
	return;
    }
#if WORDS_BIGENDIAN
    for (int i = 0; i < head.cData; ++i)
    {
	be2lel(&datap[i].type);
	be2lel(&datap[i].index);
	be2lel(&datap[i].data);
    }
#endif
	    
    tagp = new nwn_itp_tag[head.cTag];
    if (!tagp)
    {
	fprintf(stderr, "%s: Can't allocate %ld tags!\n",
		MyName, head.cTag);
	return;
    }
    fseek(itpFile, offset + head.oTag, SEEK_SET);
    if (fread(tagp, sizeof *tagp, head.cTag, itpFile) != head.cTag)
    {
	fprintf(stderr, "%s: Can't read %ld tags!\n", MyName, head.cTag);
	return;
    }
	    
    valuep = new char[head.cValue];
    if (!valuep)
    {
	fprintf(stderr, "%s: Can't allocate %ld value bytes!\n",
		MyName, head.cValue);
	return;
    }
    fseek(itpFile, offset + head.oValue, SEEK_SET);
    if (fread(valuep, sizeof *valuep, head.cValue, itpFile) != head.cValue)
    {
	fprintf(stderr, "%s: Can't read %ld value bytes!\n",
		MyName, head.cValue);
	return;
    }
	    
    /* cRoot is bytes; we want ints */
    int index = head.cRoot / sizeof *rootp;
    rootp = new int[index];
    if (!rootp)
    {
	fprintf(stderr, "%s: Can't allocate %d roots!\n", MyName, index);
	return;
    }
    fseek(itpFile, offset + head.oRoot, SEEK_SET);
    if (fread(rootp, sizeof *rootp, index, itpFile) != index)
    {
	fprintf(stderr, "%s: Can't read %d roots!\n", MyName, index);
	return;
    }
#if WORDS_BIGENDIAN
    for (int i = 0; i < (head.cRoot / sizeof *rootp); ++i)
	be2lei(&rootp[i]);
#endif
	    
    /* cIndex is bytes; we want ints */
    index = head.cIndex / sizeof *indexp;
    indexp = new int[index];
    if (!indexp)
    {
	fprintf(stderr, "%s: Can't allocate %d indexes!\n", MyName, index);
	return;
    }
    fseek(itpFile, offset + head.oIndex, SEEK_SET);
    if (fread(indexp, sizeof *indexp, index, itpFile) != index)
    {
	fprintf(stderr, "%s: Can't read %d indexes!\n", MyName, index);
	return;
    }
#if WORDS_BIGENDIAN
    for (int i = 0; i < (head.cIndex / sizeof *indexp); ++i)
	be2lei(&indexp[i]);
#endif	    
    makeMap();
    fclose(itpFile);
}


void ITPResource::makeMap() {
    if (listp[0].count)
    {
	if (listp[0].count == 1)
	    mapItem(listp[0].index / sizeof (int));
	else
	{
	    int* rp = &rootp[listp[0].index / sizeof (int)];
	    int k;
	    
	    for (k = 0; k < listp[0].count; k++) {
		mapItem(rp[k]);
	    }
	}
    }
}

void ITPResource::mapItem(int i) {
    struct nwn_itp_data* dp = &datap[i];
    int j;
    int len;
    char* s;
    
    len = 0;
    for (s = tagp[dp->index].name, j = 0; j < 16 && s[j]; ++j)
	;
    len = j;
    nameToIndexMap[std::string(s,len)] = i;
}

std::string ITPResource::getDescription() const {
    std::string desc = getName();
    if(desc.size() == 0) {
	desc = getTag();
    }
    if(desc.size() == 0) {
	desc = getResRef();
    }
    return desc;
}

std::string ITPResource::getName() const {
    return getItemByName("Name").stringValue;
}

std::string ITPResource::getResRef() const {
    return getItemByName("ResRef").stringValue;
}

std::string ITPResource::getTag() const {
    return getItemByName("Tag").stringValue;
}

NWNDataType ITPResource::getItemByName(std::string name) const {
    NWNDataType item;
    std::map<std::string,int>::const_iterator itemIndex = nameToIndexMap.find(name);
    if(itemIndex == nameToIndexMap.end()) {
	return item;
    } else {
	return getItemByIndex(itemIndex->second);
    }
}

NWNDataType ITPResource::getItemByIndex(int index) const {
    NWNDataType item;
    struct nwn_itp_data* dp = &datap[index];
    item.type = (nwn_itp_typecode)dp->type;
    int len = 0;
    int j = 0;
    char *s;
    for (s = tagp[dp->index].name, j = 0; j < 16 && s[j]; ++j)
	;
    len = j;

    item.name = std::string(s,len);    
    switch (item.type) {
	case t_uint8:
	    item.scalarValue.uint8Value = (unsigned char)dp->data;
	    break;
	case t_uint16:
	    item.scalarValue.uint16Value = (unsigned short) dp->data;
	    break;
	case t_uint32: {
	    unsigned int uii = (unsigned int) dp->data;
	    item.scalarValue.uint32Value = uii;
	    if (nw_inp_dir)
	    {
		if (item.name == "STRREF") {
		    char* cp = get_strref_str(dp->data);
		    
		    if (cp)
		    {
			item.isStringRef = true;
			item.stringValue = cp;
			delete cp;
		    }
		}
	    }
	    break;	    
	}
	case t_uint64: {
	    unsigned int uii = (unsigned int) dp->data;
	    unsigned long long ull = *((unsigned long long*) &valuep[uii]);
#if WORDS_BIGENDIAN
	    be2leull(&ull);
#endif
	    item.scalarValue.uint64Value = ull;
	    break;
	}
	case t_int8:
	    item.scalarValue.int16Value = (char) dp->data;
	    break;
	case t_int16:
	    item.scalarValue.int16Value = (short) dp->data;
	    break;
	case t_int32:
	    item.scalarValue.int32Value = (long) dp->data;
	    break;
	case t_int64: {
	    unsigned int uii = (unsigned int) dp->data;
	    unsigned long long ll = *((unsigned long long*) &valuep[uii]);
#if WORDS_BIGENDIAN
	    be2leull(&ll);
#endif
	    item.scalarValue.int64Value = ll;
	    break;
	}
	case t_float:
	    item.scalarValue.floatValue = *((float*) &dp->data);
	    break;
	case t_double:
	    item.scalarValue.doubleValue = *((double*) &valuep[dp->data]);
	    break;
	case t_string: {
	    unsigned int leng;	    
	    int ii = (int) dp->data;
	    leng = *((unsigned int*) &valuep[ii]);
#if WORDS_BIGENDIAN
	    be2leui(&leng);
#endif
	    if (leng) {
		item.stringValue = std::string(&valuep[ii + sizeof(unsigned int)],leng);
	    }
	    break;
	}	   
	case t_resref: {
	    unsigned int leng;
	    int ii = (int) dp->data;
	    leng = *((unsigned char*) &valuep[ii]);
	    item.stringValue = std::string(&valuep[ii + sizeof(unsigned char)],leng);
	    break;
	}
	case t_locstr: {
	    struct nwn_itp_slist* slp;
	    char* cp;
	    unsigned long size;
	    unsigned long data;
	    unsigned long count;
	    
	    int ii = dp->data;
	    slp = (struct nwn_itp_slist*) &valuep[ii];
	    count = slp->count;
#if WORDS_BIGENDIAN
	    be2leul(&count);
#endif
	    cp = (char*) (slp + 1);
	    if (count){
		for (int j = 0; j < count; ++j) {
		    int lang;
		    int leng;
		    
		    lang = *((int*) cp);
#if WORDS_BIGENDIAN
		    be2lei(&lang);
#endif
		    cp += sizeof (int);
		    leng = *((int*) cp);
#if WORDS_BIGENDIAN
		    be2lei(&leng);
#endif
		    cp += sizeof (int);
		    if (leng) {
			item.stringValue = std::string(cp,leng);
		    }
		    cp += leng;
		}
	    }
	    break;
	}
	case t_datref: {
	    printf("t_datref currently unhandled!\n");
	    break;
	}
	case t_capref: {
	    printf("t_capref currently unhandled!\n");
	    break;
	}
	case t_list: {
	    int ii = (int) dp->data;
	    int *ip = &indexp[ii / sizeof (int)];
	    int leng = *ip++;
	    if (leng) {
		item.listValue.clear();
		//printf("making list of %d items called %s\n", leng, item.name.c_str());
		for (int j = 0; j < leng; ++j) {
		    struct nwn_itp_list* lp;
		    int* rp;
		    int k;
		    ii = ip[j];
		    lp = &listp[ii];
		    rp = &rootp[lp->index / sizeof (int)];
		    NWNDataType listItem(t_list);
		    if (lp->count) {
			if (lp->count == 1) {
			    NWNDataType subListItem = getItemByIndex(lp->index);
			    listItem.listValue[subListItem.name] = subListItem;
			} else {
			    for (int k = 0; k < lp->count; k++) {
				NWNDataType subListItem = getItemByIndex(rp[k]);
				listItem.listValue[subListItem.name] = subListItem;
			    }
			}
		    }
		    std::ostringstream ostr;
		    ostr <<listItem.getReadableName() <<" - " <<(j+1);
		    listItem.name = listItem.getReadableName();
		    item.listValue[ostr.str()] = listItem;
		}
	    }
	    break;
	}	
	default: {
	    fprintf(stderr,"error, unhandled type in mapping\n");
	    break;
	}
    }
    return item;
}
