import os,os.path,sys
import KeyFile
import ERFFile,GFFFile,Script
import cStringIO
from sets import Set
from Progressor import Progressor

RESOURCETYPES = {
    'INVALID' : 0xFFFF,
    'BMP'     : 1,
    'TGA'     : 3,
    'WAV'     : 4,
    'PLT'     : 6,
    'INI'     : 7,
    'TXT'     : 10,
    'MDL'     : 2002,
    'NSS'     : 2009,
    'NCS'     : 2010,
    'ARE'     : 2012,
    'SET'     : 2013,
    'IFO'     : 2014,
    'BIC'     : 2015,
    'WOK'     : 2016,
    '2DA'     : 2017,
    'TXI'     : 2022,
    'GIT'     : 2023,
    'UTI'     : 2025,
    'UTC'     : 2027,
    'DLG'     : 2029,
    'ITP'     : 2030,
    'UTT'     : 2032,
    'DDS'     : 2033,
    'UTS'     : 2035,
    'LTR'     : 2036,
    'GFF'     : 2037,
    'FAC'     : 2038,
    'UTE'     : 2040,
    'UTD'     : 2042,
    'UTP'     : 2044,
    'DFT'     : 2045,
    'GIC'     : 2046,
    'GUI'     : 2047,
    'UTM'     : 2051,
    'DWK'     : 2052,
    'PWK'     : 2053,
    'JRL'     : 2056,
    'UTW'     : 2058,
    'SSF'     : 2060,
    'NDB'     : 2064,
    'PTM'     : 2065,
    'PTT'     : 2066}

RESOURCEEXTENSIONS = {}
for e in RESOURCETYPES:
    RESOURCEEXTENSIONS[RESOURCETYPES[e]] = e

GFFFILETYPES = ['ARE','IFO','BIC','GIT','UTI','UTC','DLG','ITP','UTT',\
                'UTS','GFF','FAC','UTE','UTD','UTP','GIC','GUI','UTM',\
                'JRL','UTW','PTM','PTT']

def resTypeFromExtension(ext):
    return RESOURCETYPES[ext]

def extensionFromResType(type):
    #obviously slower, but not used much in performance critical spots
    #could make reverse table if a problem
    try:
        return RESOURCEEXTENSIONS[type]
    except KeyError:
        print 'ResourceManager: type',type,'unknown resource type'
        return 'UNK'

def nameFromKey(key):
    return key[0].strip('\0') + '.' + extensionFromResType(key[1])

def keyFromName(name):
    parts = name.split('.')
    k1 = parts[0] + (16-len(parts[0]))*'\0'
    k2 = resTypeFromExtension(parts[1])
    return (k1,k2)

def interpretResourceContents(key,data):
    #could alternatively go by 'type' header field for many types...
    if extensionFromResType(key[1]) in GFFFILETYPES:
        gff = GFFFile.GFFFile()
        gff.fromFile(cStringIO.StringIO(data),0)
        return gff
    elif extensionFromResType(key[1]) == 'NSS':
        #need to do processing on script?
        return Script.Script(key[0],data)
    else:            
        return data

def keyToString(key):
    return key[0].strip('\0') + '.' + extensionFromResType(key[1])

class ResourceManager(Progressor):
    appDir = ''
    def __init__(self):
        Progressor.__init__(self)
        self.clear()

    def clear(self):
        self.keyResourceKeys = {}
        self.dirResourceKeys = {}
        self.hakResourceKeys = {}
        self.resourcesByType = {}
        self.resourcesByName = {}
        self.hakFileNames = []
        
    def scanGameDir(self,dir):
        self.clear()
        self.setAppDir(dir)
        if not os.path.exists(self.getAppDir()):
            print 'warning: dir given to resource manager does not exist'
        print "ResourceManager scanning",self.getAppDir()
        c = 0.0
        files = os.listdir(self.getAppDir())
        for f in files:
            c += 1
            self.setProgress((c/len(files))*100.0)
            if f[-4:] == '.key':
                print 'adding key file',f
                self.addKeyFile(os.path.join(self.getAppDir(),f))
            if f == 'override':
                for over in os.listdir(os.path.join(self.getAppDir(),f)):
                    pass #should add override files here
            if f == 'hak':
                for hak in os.listdir(os.path.join(self.getAppDir(),f)):
                    if hak[-4:] == '.hak':
                        #                        print 'adding hak file',hak
                        #                        self.addHAKFile(os.path.join(self.getAppDir(),f,hak))
                        self.hakFileNames.append(hak)
        self.setProgress(0)
        #actually, maybe we don't add these
        #            if f == 'modules':
        #                for mod in os.listdir(os.path.join(self.getAppDir(),f)):
        #                    if mod[-4:] == '.mod':
        #                        print 'adding mod file',mod
        #                        self.addMODFile(os.path.join(self.getAppDir(),f,mod))

    def getHAKFileNames(self):
        return self.hakFileNames

    def addKeyFile(self,keyf):
        keyFile = KeyFile.KeyFile(self.getAppDir())
        keyFile.fromFile(keyf)
        keyList = keyFile.getKeyList()
        for key in keyList:
            self.keyResourceKeys[key] = keyFile
            if key[1] in self.resourcesByType:
                self.resourcesByType[key[1]].add(key)
            else:
                self.resourcesByType[key[1]] = Set([key])

    def addHAKFile(self,hak):
        hakFile = ERFFile.ERFFile()
        hakFile.fromFile(hak)
        keyList = hakFile.getKeyList()
        for key in keyList:
            self.hakResourceKeys[key] = hakFile
            if key[1] in self.resourcesByType:
                self.resourcesByType[key[1]].add(key)
            else:
                self.resourcesByType[key[1]] = Set([key])


    def addMODFile(self,erf):
        modFile = ERFFile.ERFFile()
        modFile.fromFile(erf)
        keyList = modFile.getKeyList()
        for key in keyList:
            self.dirResourceKeys[key] = modFile
            if key[1] in self.resourcesByType:
                self.resourcesByType[key[1]].add(key)
            else:
                self.resourcesByType[key[1]] = Set([key])


    def getResource(self,key):
        if key in self.hakResourceKeys:
            return self.hakResourceKeys[key].getResource(key)
        elif key in self.dirResourceKeys:
            return self.dirResourceKeys[key].getResource(key)
        elif key in self.keyResourceKeys:
            return self.keyResourceKeys[key].getResource(key)
        else:
            print 'error: cannot find resource for ',key,'in any added lists'

    def getKeysWithExtensions(self,ext):
        return self.resourcesByType[resTypeFromExtension(ext)]

    def getTextScriptKeys(self):
        return self.getKeysWithExtensions('nss')

    def getAppDir(cls):
        return cls.appDir
    getAppDir = classmethod(getAppDir)

    def setAppDir(cls,dir):
        cls.appDir = dir
    setAppDir = classmethod(setAppDir)
